#include <string>
using namespace std;
/*
	xpm2rule.c++
	blah 2017-2018

	This is a program which takes an xpm image as input, and creates a rule to
	be run in Golly. It does not support xpm files in which each pixel is
	represented by 2 characters; the way GIMP seems to generate them, this means
	that the input must have no more than 91 unique colours.

	I have only tested it for 90, but the limit should be 91.

	To use it, you must first generate an image with a limited colour palette.
	This can be done in GIMP by going to "Image>Mode>Indexed..." (assuming your
	image does not already have a limited colour palette), and I'm sure you can
	figure it out from there.

	The image can't be too big, since it has to be generated within 32^4=1048576
	generations. Roughly speaking, each pixel takes 4 generations to produce, so
	your image must have less than a quarter of a million pixels.

	You can then export it as an XPM file. Once you have done this, open the XPM
	file in a text editor and modify the text around the beginning to change
	"static char *" into "string". This is necessary because I'm lazu. I could
	fix that typo, as well, but I'm too lazu.

	(I have just noticed, in trying to test this procedure, that it's possible
	for the XPM to be generated without a colour called "None". Good luck.)

	Once you have the .xpm file, modify the following lines to refer to your
	file:
*/
#include "lol.xpm"
#define xpmin lol_xpm
/*
	The filename should be in the include line, and the second line should use
	the text between "string " and "[]".

	This program can then be compiled with this command:
	g++ xpm2rule.c++ -std=c++11
	This assumes you have GCC and a unix-like command terminal. You could of
	course use an IDE or any other means; all that matters is that it's compiled
	with C++11.

	You can then run the program; it will output a file called "rom.txt", which
	you can then use as a rule in Golly. If it works, wow, somebody else gave
	enough of a fuck and got lucky enough to actually make this work.

	Did I miss anything? Probably not.
*/

#include <iostream>
typedef long num;
#include <fstream> // for outputting

const string ovtbp="11";
const string ovtp="12";
const string ohtp="13";
const string o="1";


num t=0;
// Counters
string C(){
	// From least to most significant digit
	string d1 = to_string((t%32)+14);
	string d2 = to_string(((t>>5)%32)+14);
	string d3 = to_string(((t>>10)%32)+14);
	string d4 = to_string(((t>>15)%32)+14);
	return "\nO,"+d3+","+d4+",d,e,f,g,"+d1+","+d2+",";
}

// Used to convert hexadecimal colours in xpm to decimal.
// Lovingly plagiarised from my own earlier work
unsigned char charToNum(char in){
	if(in>='0' && in<='9')
		return in-48;
	else if(in>='A' && in<='F')
		return in-55;
	else if(in>='a' && in<='f')
		return in-87;
	return 0;
}
string hexToDec(string in){
	return to_string(charToNum(in[1]) | (charToNum(in[0]) << 4));
}

int main(){
	int xpmWidth = 200;
	int xpmHeight = 298;
	int xpmColours = 13;
	{ //Interpret metadata (width, height, colours)
		int spaceCount=0;
		for(int i=xpmin[0].length();i>=0;i--){
			if(xpmin[0][i]==' '){
				if(spaceCount==1){
					xpmColours=stoi(xpmin[0].substr(i+1))-1;
				}else if(spaceCount==2){
					xpmHeight=stoi(xpmin[0].substr(i+1));
				}
				spaceCount++;
				xpmin[0]=xpmin[0].substr(0,i);
			}
		}
		xpmWidth=stoi(xpmin[0]);
	}
	char xpmCharIn[104] =   {};
	string xpmNumOut[104] = {};

	for(int i=0;i<xpmColours;i++){
		xpmCharIn[i]=xpmin[2+i][0];
		xpmNumOut[i]=to_string(47+i);
	}

	string ret="";
	ret+=C()+ovtbp; t++;
	ret+=C()+o; t++;
	for(num y=0;y<xpmHeight;y++){
		ret+=C()+ovtp; t++;
		ret+=C()+o; t++;
		ret+=C()+ohtp; t+=xpmWidth*2-2;
		int lastChar = -1; //-1 will never be equal to anything in the real world, dumbass
		for(num x=0;x<xpmWidth;x++){
			for(int i=0;i<xpmColours;i++){
				if(xpmin[2+xpmColours+y][xpmWidth-x-1]==xpmCharIn[i]){
					if(lastChar!=i) // If the last character was the same, there is no reason to change the state to generate it; it is already being generated.
						ret+=C()+xpmNumOut[i];
					t+=2; lastChar=i;
				}
			}
		}
		ret+=C()+o; t++; // "o;" looks like a face.
	}
	ret+=C()+"ki";

	string stitchColours="";
	for(int i=0;i<xpmColours;i++){
		stitchColours+=to_string(47+xpmColours+i)+" "+
		hexToDec(xpmin[2+i].substr(5,2))+" "+
		hexToDec(xpmin[2+i].substr(7,2))+" "+
		hexToDec(xpmin[2+i].substr(9,2))+"\n";
	}
	string stitchVars="var a={0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46";
	for(int i=0;i<xpmColours*2;i++){
		stitchVars+=","+to_string(47+i);
	}
	stitchVars+="}\nvar A={";
	for(int i=0;i<xpmColours;i++){
		stitchVars+=(i==0?"":",")+to_string(47+xpmColours+i);
	}
	stitchVars+="}\nvar O={1,11,12,13";
	for(int i=0;i<xpmColours;i++){
		stitchVars+=","+to_string(47+i);
	}
	stitchVars+="}";
	string stitchSpawns="";
	for(int i=0;i<xpmColours;i++){
		stitchSpawns+="a,b,c,d,e,f,g,h,"+to_string(47+i)+","+to_string(47+xpmColours+i)+"\n";
	}

	// Output file
	cout<<"Done.\n" <<xpmWidth<<" "<<xpmHeight<<" "<<xpmColours<<"\n"<<stitchSpawns<<"\n";
	ofstream outfile;
	outfile.open("rom.txt");
	outfile << R".zoi.(@RULE BLAHPRINTBASE
2017-2018
This is a rule which was automatically generated by xpm2rule.c++, by blah.
It will print out an image using means detailed here:
http://conwaylife.com/forums/viewtopic.php?f=11&t=3088


@NAMES
0 empty (e)
1 source (o)
2 horizontal target (ht)
3 ht pusher (htp)
4 ht being pushed (htbp)
5 horizontal sheath (hs)
6 vertical target (vt)
7 vt pusher (vtp)
8 vt being pushed (vtbp)
9 vertical sheath (vs)
10 kill umbilical cord (ki)
# commands; states the origin takes to do things
11 spawn vtbp (ovtbp)
12 spawn vtp (ovtp)
13 spawn htp (ohtp)
# counter states are 14 to 45
46 digit carry (dc)
# all states above are paint/paint generators

@COLORS

0 0 0 0
).zoi."<< stitchColours <<R".zoi.(

@TABLE

n_states:).zoi."<< 47+xpmColours*2 <<R".zoi.(
neighborhood:Moore
symmetries:none
).zoi."<< stitchVars <<R".zoi.(
var b=a
var c=a
var d=a
var e=a
var f=a
var g=a
var h=a
var i=a
var Ca={14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45}
var Cb={14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45}
var Cc={14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45}
var ARM={6,9,7,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45}

var ht={2}
var htp={3}
var htbp={4}
var hs={5}
var vt={6}
var vtp={7}
var vtbp={8}
var vs={9}
var ki={10}
var ovtbp={11}
var ovtp={12}
var ohtp={13}
var dc={46}


# A = colours
# O = origins
# Cx = counters
# the other vars are all states


# COUNTER TRANSITIONS BEGIN
# we get counters around the origin
0,b,c,d,e,f,1,h,i,14
0,b,c,d,e,1,g,h,i,14
0,b,c,d,1,f,g,h,i,14
0,b,c,1,e,f,g,h,i,14
# digit 1 (lsd)
14,b,c,O,e,f,g,h,i,15
15,b,c,O,e,f,g,h,i,16
16,b,c,O,e,f,g,h,i,17
17,b,c,O,e,f,g,h,i,18
18,b,c,O,e,f,g,h,i,19
19,b,c,O,e,f,g,h,i,20
20,b,c,O,e,f,g,h,i,21
21,b,c,O,e,f,g,h,i,22
22,b,c,O,e,f,g,h,i,23
23,b,c,O,e,f,g,h,i,24
24,b,c,O,e,f,g,h,i,25
25,b,c,O,e,f,g,h,i,26
26,b,c,O,e,f,g,h,i,27
27,b,c,O,e,f,g,h,i,28
28,b,c,O,e,f,g,h,i,29
29,b,c,O,e,f,g,h,i,30
30,b,c,O,e,f,g,h,i,31
31,b,c,O,e,f,g,h,i,32
32,b,c,O,e,f,g,h,i,33
33,b,c,O,e,f,g,h,i,34
34,b,c,O,e,f,g,h,i,35
35,b,c,O,e,f,g,h,i,36
36,b,c,O,e,f,g,h,i,37
37,b,c,O,e,f,g,h,i,38
38,b,c,O,e,f,g,h,i,39
39,b,c,O,e,f,g,h,i,40
40,b,c,O,e,f,g,h,i,41
41,b,c,O,e,f,g,h,i,42
42,b,c,O,e,f,g,h,i,43
43,b,c,O,e,f,g,h,i,44
44,b,c,O,e,f,g,h,i,45
45,b,c,O,e,f,g,h,i,14
# digit 2
14,b,c,d,O,45,g,h,i,15
15,b,c,d,O,45,g,h,i,16
16,b,c,d,O,45,g,h,i,17
17,b,c,d,O,45,g,h,i,18
18,b,c,d,O,45,g,h,i,19
19,b,c,d,O,45,g,h,i,20
20,b,c,d,O,45,g,h,i,21
21,b,c,d,O,45,g,h,i,22
22,b,c,d,O,45,g,h,i,23
23,b,c,d,O,45,g,h,i,24
24,b,c,d,O,45,g,h,i,25
25,b,c,d,O,45,g,h,i,26
26,b,c,d,O,45,g,h,i,27
27,b,c,d,O,45,g,h,i,28
28,b,c,d,O,45,g,h,i,29
29,b,c,d,O,45,g,h,i,30
30,b,c,d,O,45,g,h,i,31
31,b,c,d,O,45,g,h,i,32
32,b,c,d,O,45,g,h,i,33
33,b,c,d,O,45,g,h,i,34
34,b,c,d,O,45,g,h,i,35
35,b,c,d,O,45,g,h,i,36
36,b,c,d,O,45,g,h,i,37
37,b,c,d,O,45,g,h,i,38
38,b,c,d,O,45,g,h,i,39
39,b,c,d,O,45,g,h,i,40
40,b,c,d,O,45,g,h,i,41
41,b,c,d,O,45,g,h,i,42
42,b,c,d,O,45,g,h,i,43
43,b,c,d,O,45,g,h,i,44
44,b,c,d,O,45,g,h,i,45
45,b,c,d,O,45,g,h,i,14
# digit 3
14,b,c,d,e,O,45,45,i,15
15,b,c,d,e,O,45,45,i,16
16,b,c,d,e,O,45,45,i,17
17,b,c,d,e,O,45,45,i,18
18,b,c,d,e,O,45,45,i,19
19,b,c,d,e,O,45,45,i,20
20,b,c,d,e,O,45,45,i,21
21,b,c,d,e,O,45,45,i,22
22,b,c,d,e,O,45,45,i,23
23,b,c,d,e,O,45,45,i,24
24,b,c,d,e,O,45,45,i,25
25,b,c,d,e,O,45,45,i,26
26,b,c,d,e,O,45,45,i,27
27,b,c,d,e,O,45,45,i,28
28,b,c,d,e,O,45,45,i,29
29,b,c,d,e,O,45,45,i,30
30,b,c,d,e,O,45,45,i,31
31,b,c,d,e,O,45,45,i,32
32,b,c,d,e,O,45,45,i,33
33,b,c,d,e,O,45,45,i,34
34,b,c,d,e,O,45,45,i,35
35,b,c,d,e,O,45,45,i,36
36,b,c,d,e,O,45,45,i,37
37,b,c,d,e,O,45,45,i,38
38,b,c,d,e,O,45,45,i,39
39,b,c,d,e,O,45,45,i,40
40,b,c,d,e,O,45,45,i,41
41,b,c,d,e,O,45,45,i,42
42,b,c,d,e,O,45,45,i,43
43,b,c,d,e,O,45,45,i,44
44,b,c,d,e,O,45,45,i,45
45,b,c,d,e,O,45,45,i,14
# my finger hurts
# dc is spawned
0,0,0,45,42,0,0,0,0,dc
# dc moves
dc,b,c,d,e,f,g,h,i,0
0,b,c,d,45,f,dc,h,i,dc
0,b,c,d,Ca,Cb,Cc,dc,i,dc
# digit 4
14,b,c,d,e,f,O,h,dc,15
15,b,c,d,e,f,O,h,dc,16
16,b,c,d,e,f,O,h,dc,17
17,b,c,d,e,f,O,h,dc,18
18,b,c,d,e,f,O,h,dc,19
19,b,c,d,e,f,O,h,dc,20
20,b,c,d,e,f,O,h,dc,21
21,b,c,d,e,f,O,h,dc,22
22,b,c,d,e,f,O,h,dc,23
23,b,c,d,e,f,O,h,dc,24
24,b,c,d,e,f,O,h,dc,25
25,b,c,d,e,f,O,h,dc,26
26,b,c,d,e,f,O,h,dc,27
27,b,c,d,e,f,O,h,dc,28
28,b,c,d,e,f,O,h,dc,29
29,b,c,d,e,f,O,h,dc,30
30,b,c,d,e,f,O,h,dc,31
31,b,c,d,e,f,O,h,dc,32
32,b,c,d,e,f,O,h,dc,33
33,b,c,d,e,f,O,h,dc,34
34,b,c,d,e,f,O,h,dc,35
35,b,c,d,e,f,O,h,dc,36
36,b,c,d,e,f,O,h,dc,37
37,b,c,d,e,f,O,h,dc,38
38,b,c,d,e,f,O,h,dc,39
39,b,c,d,e,f,O,h,dc,40
40,b,c,d,e,f,O,h,dc,41
41,b,c,d,e,f,O,h,dc,42
42,b,c,d,e,f,O,h,dc,43
43,b,c,d,e,f,O,h,dc,44
44,b,c,d,e,f,O,h,dc,45
45,b,c,d,e,f,O,h,dc,14

# COUNTER TRANSITIONS END

# origin spawns things to the lower right
a,b,c,d,e,f,g,h,ovtbp,vtbp
a,b,c,d,e,f,g,h,ovtp,vtp
a,b,c,d,e,f,g,h,ohtp,htp
).zoi."<<stitchSpawns<<R".zoi.(


# htp moves down next to vs
0,htp,c,d,e,f,g,vs,i,htp
htp,b,c,d,e,f,g,h,i,0
# htp moves to the right over hs
0,b,c,d,e,hs,g,htp,i,htp
htp,b,c,d,e,hs,g,h,i,0

# ht is moved to the right by htp
ht,b,c,d,e,f,g,htp,i,htbp
htbp,b,c,d,e,f,g,h,i,0
0,b,c,d,e,f,g,htbp,i,ht
# leaving a sheath of hs
0,htbp,c,d,e,f,g,h,i,hs

# vtp moves down
0,vtp,c,d,e,f,g,h,i,vtp
vtp,b,c,d,e,f,g,h,i,0
# vt is moved down by vtp
vt,vtp,c,d,e,f,g,h,i,vtbp
vtbp,b,c,d,e,f,g,h,i,0
0,vtbp,c,d,e,f,g,h,i,vt
0,b,c,vtbp,e,f,g,h,i,vs

# htp spawns horizontal arm at vt
0,b,c,d,e,f,vt,htp,i,htbp

# data places itself where it goes
ht,b,c,d,e,f,g,A,i,A
A,b,c,ht,e,f,g,h,i,ht
hs,A,ht,d,e,f,g,h,i,0
ht,b,c,d,e,vt,g,h,i,0
# data moves down
0,A,c,d,e,f,g,vs,i,A
A,b,c,d,e,f,g,vs,i,0
# and to the right
0,b,c,d,e,hs,g,A,i,A
A,b,c,d,e,hs,g,h,i,0

#kill kills
ki,b,c,d,e,f,g,h,i,0
ARM,ki,c,d,e,f,g,h,i,ki
ARM,b,ki,d,e,f,g,h,i,ki
ARM,b,c,ki,e,f,g,h,i,ki
ARM,b,c,d,ki,f,g,h,i,ki
ARM,b,c,d,e,ki,g,h,i,ki
ARM,b,c,d,e,f,ki,h,i,ki
ARM,b,c,d,e,f,g,ki,i,ki
ARM,b,c,d,e,f,g,h,ki,ki

# ~=~=~ ROM ~=~=~
).zoi." << ret << '\n';
	outfile.close();
}
